from collections.abc import Mapping as _Mapping
from contextlib import ExitStack
from copy import copy
from os import PathLike
from typing import Any, Optional, Union, Mapping, Dict
from warnings import warn

import jaydebeapi
from jpype import JVMNotFoundException, isJVMStarted, startJVM, getDefaultJVMPath
from pandakeeper.dataloader.sql import SqlLoader
from pandakeeper.validators import AnyDataFrame
from pandera import DataFrameSchema
from typing_extensions import final
from varutils.plugs.constants import empty_mapping_proxy
from varutils.typing import check_type_compatibility

from sber_ld_dbtools.credentials import PasswordKeeper

__all__ = (
    'OracleContextManager',
    'OracleLoader',
    'GlobalOracleConfig'
)


class OracleContextManager:
    __slots__ = ('host', 'user', '__password', 'port', 'sid', 'connection')

    def __init__(self, host: str, *, user: str, password: str, port: Union[str, int], sid: str) -> None:

        check_type_compatibility(host, str)
        check_type_compatibility(user, str)
        check_type_compatibility(password, str)
        check_type_compatibility(port, (str, int))
        check_type_compatibility(sid, str)

        self.host = host
        self.user = user
        self.__password = password
        self.port = port
        self.sid = sid
        self.connection: Optional[jaydebeapi.Connection] = None

    def __enter__(self) -> jaydebeapi.Connection:
        if not isJVMStarted():
            startJVM(
                GlobalOracleConfig.JVMPath,
                f'-Djava.class.path={GlobalOracleConfig.OJDBC8_PATH}',
                convertStrings=True
            )
        self.connection = jaydebeapi.connect(
            'oracle.jdbc.driver.OracleDriver',
            f'jdbc:oracle:thin:{self.user}/{self.__password}@{self.host}:{self.port}:{self.sid}'
        )
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self.connection is not None:
            self.connection.close()
            self.connection = None
        if exc_type is not None:
            raise


def _oracle_context_creator(stack: ExitStack,
                            credentials: PasswordKeeper,
                            **oracle_parameters: Any) -> jaydebeapi.Connection:
    user = oracle_parameters.get('user')
    password = oracle_parameters.get('password')
    if user is None:
        user = credentials.get_username()
    if password is None:
        password = credentials.get_password()
    host = oracle_parameters['host']
    port = oracle_parameters['port']
    sid = oracle_parameters['sid']

    conn = stack.enter_context(
        OracleContextManager(
            host,
            user=user,
            password=password,
            port=port,
            sid=sid
        )
    )
    return conn


class OracleLoader(SqlLoader):
    __slots__ = ()

    def __init__(self,
                 sql_query: str,
                 *,
                 credentials: PasswordKeeper,
                 oracle_parameters: Mapping[str, Any] = empty_mapping_proxy,
                 output_validator: DataFrameSchema = AnyDataFrame,
                 **read_sql_kwargs: Any) -> None:

        check_type_compatibility(credentials, PasswordKeeper)
        check_type_compatibility(oracle_parameters, _Mapping, 'Mapping')

        for keyword in ('host', 'port', 'sid'):
            if keyword not in oracle_parameters:
                raise KeyError(f"Parameter 'oracle_parameters' should contain '{keyword}' key")

        super().__init__(
            _oracle_context_creator,
            sql_query,
            context_creator_args=(credentials,),
            context_creator_kwargs=copy(oracle_parameters),
            read_sql_kwargs=read_sql_kwargs,
            output_validator=output_validator
        )

    @final
    @property
    def credentials(self) -> PasswordKeeper:
        return self._context_creator_args[0]

    @final
    @property
    def oracle_parameters(self) -> Dict[str, Any]:
        res = dict(self._context_creator_kwargs)
        if 'password' in res:
            res['password'] = '*****'
        return res


class _GlobalOracleConfigType:
    __slots__ = ()
    __instance: Optional['_GlobalOracleConfigType'] = None
    __OJDBC8_PATH = ''
    try:
        __JVMPath: str = getDefaultJVMPath()
    except JVMNotFoundException:
        warn(
            "No JVM shared library file (libjli.dylib) found. "
            "Try setting up the JAVA_HOME environment variable properly "
            "or setting up the GlobalOracleConfig.JVMPath property manually.",
            RuntimeWarning
        )
        __JVMPath = ''

    def __new__(cls) -> '_GlobalOracleConfigType':
        instance = _GlobalOracleConfigType.__instance
        if instance is None:
            instance = super().__new__(cls)
            _GlobalOracleConfigType.__instance = instance
        return instance

    @property
    def OJDBC8_PATH(self) -> str:
        res = _GlobalOracleConfigType.__OJDBC8_PATH
        if not res:
            raise ValueError('OJDBC8_PATH is not set')
        return res

    @OJDBC8_PATH.setter
    def OJDBC8_PATH(self, value: Union[str, bytes, PathLike]) -> None:
        check_type_compatibility(value, (str, bytes, PathLike))
        _GlobalOracleConfigType.__OJDBC8_PATH = str(value)

    @property
    def JVMPath(self) -> str:
        return _GlobalOracleConfigType.__JVMPath

    @JVMPath.setter
    def JVMPath(self, value: Union[str, bytes, PathLike]) -> None:
        check_type_compatibility(value, (str, bytes, PathLike))
        _GlobalOracleConfigType.__JVMPath = str(value)


GlobalOracleConfig = _GlobalOracleConfigType()



from collections.abc import Mapping as _Mapping
from contextlib import ExitStack
from copy import copy
from typing import Any, Optional, Mapping, Dict

import teradatasql
from pandakeeper.dataloader.sql import SqlLoader
from pandakeeper.validators import AnyDataFrame
from pandera import DataFrameSchema
from typing_extensions import final
from varutils.plugs.constants import empty_mapping_proxy
from varutils.typing import check_type_compatibility

from sber_ld_dbtools.credentials import PasswordKeeper, set_default_kerberos_principal

__all__ = (
    'TeradataContextManager',
    'TeradataLoader'
)


class TeradataContextManager:
    __slots__ = ('host', 'user', '__password', 'logmech', 'connection')

    def __init__(self, host: str, *, user: str, password: str, logmech: str) -> None:

        check_type_compatibility(host, str)
        check_type_compatibility(user, str)
        check_type_compatibility(password, str)
        check_type_compatibility(logmech, str)

        self.host = host
        self.user = user
        self.__password = password
        self.logmech = logmech
        self.connection: Optional[teradatasql.TeradataConnection] = None

    def __enter__(self) -> teradatasql.TeradataConnection:
        self.connection = teradatasql.connect(
            f'{{"host": "{self.host}"}}',
            user=self.user,
            password=self.__password,
            logmech=self.logmech
        )
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self.connection is not None:
            self.connection.close()
            self.connection = None
        if exc_type is not None:
            raise


def _teradata_context_creator(stack: ExitStack,
                              credentials: PasswordKeeper,
                              **teradata_context_kwargs: str) -> teradatasql.TeradataConnection:
    set_default_kerberos_principal(credentials)
    if 'user' not in teradata_context_kwargs:
        teradata_context_kwargs['user'] = credentials.get_username()
    if 'password' not in teradata_context_kwargs:
        teradata_context_kwargs['password'] = credentials.get_password()
    conn = stack.enter_context(TeradataContextManager(**teradata_context_kwargs))
    return conn


class TeradataLoader(SqlLoader):
    __slots__ = ()

    def __init__(self,
                 sql_query: str,
                 *,
                 credentials: PasswordKeeper,
                 teradata_parameters: Mapping[str, Any] = empty_mapping_proxy,
                 output_validator: DataFrameSchema = AnyDataFrame,
                 **read_sql_kwargs: Any) -> None:

        check_type_compatibility(credentials, PasswordKeeper)
        check_type_compatibility(teradata_parameters, _Mapping, 'Mapping')

        for teradata_keyword in ('host', 'logmech'):
            if teradata_keyword not in teradata_parameters:
                raise KeyError(f"Parameter 'teradata_parameters' should contain '{teradata_keyword}' key")

        super().__init__(
            _teradata_context_creator,
            sql_query,
            context_creator_args=(credentials,),
            context_creator_kwargs=copy(teradata_parameters),
            read_sql_kwargs=read_sql_kwargs,
            output_validator=output_validator
        )

    @final
    @property
    def credentials(self) -> PasswordKeeper:
        return self._context_creator_args[0]

    @final
    @property
    def teradata_parameters(self) -> Dict[str, Any]:
        res = dict(self._context_creator_kwargs)
        if 'password' in res:
            res['password'] = '*****'
        return res



from collections.abc import Mapping as _Mapping
from contextlib import ExitStack
from copy import copy
from typing import Any, Mapping, Dict

import psycopg2
from pandakeeper.dataloader.sql import SqlLoader
from pandakeeper.validators import AnyDataFrame
from pandera import DataFrameSchema
from typing_extensions import final
from varutils.plugs.constants import empty_mapping_proxy
from varutils.typing import check_type_compatibility

from sber_ld_dbtools.credentials import PasswordKeeper, set_default_kerberos_principal

__all__ = (
    'GreenplumLoader',
)


def _greenplum_context_creator(stack: ExitStack,
                               credentials: PasswordKeeper,
                               **greenplum_context_kwargs: Any) -> psycopg2.extensions.connection:
    set_default_kerberos_principal(credentials)
    if 'user' not in greenplum_context_kwargs:
        greenplum_context_kwargs['user'] = credentials.get_username()
    conn = stack.enter_context(psycopg2.connect(**greenplum_context_kwargs))
    return conn


class GreenplumLoader(SqlLoader):
    __slots__ = ()

    def __init__(self,
                 sql_query: str,
                 *,
                 credentials: PasswordKeeper,
                 greenplum_parameters: Mapping[str, Any] = empty_mapping_proxy,
                 output_validator: DataFrameSchema = AnyDataFrame,
                 **read_sql_kwargs: Any) -> None:

        check_type_compatibility(credentials, PasswordKeeper)
        check_type_compatibility(greenplum_parameters, _Mapping, 'Mapping')

        for keyword in ('host', 'dbname'):
            if keyword not in greenplum_parameters:
                raise KeyError(f"Parameter 'greenplum_parameters' should contain '{keyword}' key")

        super().__init__(
            _greenplum_context_creator,
            sql_query,
            context_creator_args=(credentials,),
            context_creator_kwargs=copy(greenplum_parameters),
            read_sql_kwargs=read_sql_kwargs,
            output_validator=output_validator
        )

    @final
    @property
    def credentials(self) -> PasswordKeeper:
        return self._context_creator_args[0]

    @final
    @property
    def greenplum_parameters(self) -> Dict[str, Any]:
        res = dict(self._context_creator_kwargs)
        if 'password' in res:
            res['password'] = '*****'
        return res



from getpass import getpass
from io import StringIO
from os import PathLike
from subprocess import run as run_cmd, CalledProcessError
from typing import Optional, Union

from varutils.types import NoneType
from varutils.typing import check_type_compatibility

__all__ = (
    'PasswordKeeper',
    'KerberosPrincipalSwitcher',
    'set_default_kerberos_principal',
)


class PasswordKeeper:
    __slots__ = ('username', 'host', 'prompt', '__password')

    def __init__(self,
                 username: str,
                 *,
                 host: Optional[str] = None,
                 prompt: str = '',
                 passwd_file_or_buffer: Union[str, bytes, PathLike, StringIO, None] = None) -> None:

        check_type_compatibility(username, str)
        check_type_compatibility(host, (str, NoneType), 'str or None')
        check_type_compatibility(prompt, str)
        check_type_compatibility(passwd_file_or_buffer, (str, bytes, PathLike, StringIO, NoneType))

        self.username = username
        self.host = host
        self.prompt = prompt
        if isinstance(passwd_file_or_buffer, (str, bytes, PathLike)):
            with open(passwd_file_or_buffer, 'r') as f:
                self.__password = f.read().strip('\n')
        elif isinstance(passwd_file_or_buffer, StringIO):
            self.__password = passwd_file_or_buffer.read().strip('\n')
        else:
            self.__password = getpass(prompt=prompt)

    def set_password(self) -> 'PasswordKeeper':
        self.__password = getpass(prompt=self.prompt)
        return self

    def get_username(self) -> str:
        return self.username

    def get_host(self) -> str:
        if self.host is None:
            raise ValueError("host is not set")
        return self.host

    def get_username_with_host(self) -> str:
        return f'{self.username}@{self.get_host()}'

    def get_password(self) -> str:
        return self.__password


def set_default_kerberos_principal(passkeeper: PasswordKeeper) -> None:
    username_with_host = passkeeper.get_username_with_host()
    password = passkeeper.get_password()

    cmd_base = f'kinit {username_with_host}'
    completed_process = run_cmd(
        f"unset HISTFILE && {cmd_base} <<< '{password}'",
        shell=True,
        capture_output=True,
        text=True
    )
    if completed_process.returncode:
        raise CalledProcessError(
            completed_process.returncode,
            f'{cmd_base} <<< ***',
            completed_process.stderr.strip()
        )


class KerberosPrincipalSwitcher:
    __slots__ = ('__context_credentials', '__exit_credentials')

    def __init__(self, context_credentials: PasswordKeeper, exit_credentials: PasswordKeeper) -> None:
        check_type_compatibility(context_credentials, PasswordKeeper)
        check_type_compatibility(exit_credentials, PasswordKeeper)
        self.__context_credentials = context_credentials
        self.__exit_credentials = exit_credentials

    def __enter__(self) -> None:
        set_default_kerberos_principal(self.__context_credentials)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        set_default_kerberos_principal(self.__exit_credentials)
        if exc_type is not None:
            raise



from typing import Optional

from pandakeeper.validators import AnyDataFrame
from pandas import DataFrame
from pandera import DataFrameSchema
from pyspark import SparkContext, HiveContext, SparkConf

from sber_ld_dbtools.credentials import PasswordKeeper
from sber_ld_dbtools.loader.spark import SparkBaseLoader

__all__ = (
    'SparkHiveLoader',
)


def _read_hive_sql(sql_query: str, conn: SparkContext) -> DataFrame:
    hc = HiveContext(conn)
    sql_result = hc.sql(sql_query)
    return sql_result.toPandas()


class SparkHiveLoader(SparkBaseLoader):
    __slots__ = ()

    def __init__(self,
                 sql_query: str,
                 *,
                 credentials: PasswordKeeper,
                 conf: Optional[SparkConf] = None,
                 output_validator: DataFrameSchema = AnyDataFrame) -> None:
        super().__init__(
            sql_query,
            credentials=credentials,
            conf=conf,
            read_sql_fn=_read_hive_sql,
            output_validator=output_validator
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sber_ld_dbtools',
 'sber_ld_dbtools.credentials',
 'sber_ld_dbtools.loader',
 'sber_ld_dbtools.loader.greenplum',
 'sber_ld_dbtools.loader.oracle',
 'sber_ld_dbtools.loader.spark',
 'sber_ld_dbtools.loader.spark.hive',
 'sber_ld_dbtools.loader.teradata']

package_data = \
{'': ['*']}

install_requires = \
['JPype1>=1,<2',
 'JayDeBeApi>=1,<2',
 'pandakeeper>=0.0.24,<0.0.25',
 'pandas>=1,<2',
 'psycopg2-binary>=2,<3',
 'pyspark>=2,<3',
 'teradatasql>=17,<18',
 'varutils>=0.0.6,<0.0.7']

setup_kwargs = {
    'name': 'sber-ld-dbtools',
    'version': '0.0.6',
    'description': 'Tools for interacting with LD databases.',
    'long_description': '# sber_ld_dbtools\nTools for interacting with LD databases.\n',
    'author': 'Andrew Sonin',
    'author_email': 'sonin.cel@yandex.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/andrewsonin/sber_ld_dbtools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from contextlib import ExitStack
from os import environ, PathLike
from typing import Optional, Union, Callable, Tuple, Mapping, Any

from pandakeeper.dataloader.sql import SqlLoader
from pandas import DataFrame
from pandas import read_sql
from pandera import DataFrameSchema
from pyspark import SparkConf, SparkContext
from typing_extensions import Final, final
from varutils.plugs.constants import empty_mapping_proxy
from varutils.types import NoneType
from varutils.typing import check_type_compatibility, get_fully_qualified_name

from sber_ld_dbtools.credentials import PasswordKeeper, set_default_kerberos_principal

__all__ = (
    'SparkBaseLoader',
    'GlobalSparkConfig',
    'DEFAULT_SPARK_CONF'
)


def _spark_context_creator(stack: ExitStack,
                           credentials: PasswordKeeper,
                           conf: Optional[SparkConf] = None) -> SparkContext:
    set_default_kerberos_principal(credentials)
    return stack.enter_context(SparkContext.getOrCreate(conf))


class SparkBaseLoader(SqlLoader):
    __slots__ = ()

    def __init__(self,
                 sql_query: str,
                 *,
                 credentials: PasswordKeeper,
                 conf: Optional[SparkConf] = None,
                 read_sql_fn: Callable[..., DataFrame] = read_sql,
                 read_sql_args: Tuple[Any, ...] = (),
                 read_sql_kwargs: Mapping[str, Any] = empty_mapping_proxy,
                 output_validator: DataFrameSchema) -> None:
        check_type_compatibility(credentials, PasswordKeeper)
        check_type_compatibility(conf, (SparkConf, NoneType), f'{get_fully_qualified_name(SparkConf)} or None')

        super().__init__(
            _spark_context_creator,
            sql_query,
            context_creator_args=(credentials, conf),
            read_sql_fn=read_sql_fn,
            read_sql_args=read_sql_args,
            read_sql_kwargs=read_sql_kwargs,
            output_validator=output_validator
        )

    @final
    @property
    def credentials(self) -> PasswordKeeper:
        return self._context_creator_args[0]

    @final
    @property
    def conf(self) -> Optional[SparkConf]:
        return self._context_creator_args[1]


DEFAULT_SPARK_CONF: Final = (
    SparkConf()
        .setAppName('DownloadRawData')
        .setMaster("yarn-client")
        .set('spark.dynamicAllocation.enabled', 'false')
        .set('spark.local.dir', '.sparktmp')
        .set('spark.executor.memory', '6g')
        .set('spark.executor.cores', '2')
        .set('spark.executor.instances', '50')
        .set('spark.sql.parquet.mergeScheme', 'false')
        .set('parquet.enable.summary-metadata', 'false')
        .set('spark.yarn.executor.memoryOverhead', '6048mb')
        .set('spark.driver.memory', '90g')
        .set('spark.driver.maxResultSize', '90g')
        .set('spark.yarn.driver.memoryOverhead', '6048mb')
        .set('spark.port.maxRetries', '150')
        .set('spark.dynamicAllocation.enabled', 'false')
        .set('spark.kryoserializer.buffer.max', '1g')
        .set('spark.core.connection.ack.wait.timeout', '800s')
        .set('spark.akka.timeout', '800s')
        .set('spark.storage.blockManagerSlaveTimeoutMs', '800s')
        .set('spark.shuffle.io.connectionTimeout', '800s')
        .set('spark.rpc.askTimeout', '800s')
        .set('spark.network.timeout', '800s')
        .set('spark.rpc.lookupTimeout', '800s')
        .set('spark.sql.crossJoin.enabled', 'True')
        .set('spark.sql.autoBroadcastJoinThreshold', -1)
)


class _GlobalSparkConfigType:
    __slots__ = ()
    __instance: Optional['_GlobalSparkConfigType'] = None

    def __new__(cls) -> '_GlobalSparkConfigType':
        instance = _GlobalSparkConfigType.__instance
        if instance is None:
            instance = super().__new__(cls)
            _GlobalSparkConfigType.__instance = instance
        return instance

    @property
    def SPARK_HOME(self) -> Optional[str]:
        return environ.get('SPARK_HOME')

    @SPARK_HOME.setter
    def SPARK_HOME(self, value: Union[str, bytes, PathLike]) -> None:
        check_type_compatibility(value, (str, bytes, PathLike))
        environ['SPARK_HOME'] = str(value)

    @property
    def PYSPARK_DRIVER_PYTHON(self) -> Optional[str]:
        return environ.get('PYSPARK_DRIVER_PYTHON')

    @PYSPARK_DRIVER_PYTHON.setter
    def PYSPARK_DRIVER_PYTHON(self, value: str) -> None:
        check_type_compatibility(value, str)
        environ['PYSPARK_DRIVER_PYTHON'] = value

    @property
    def PYSPARK_PYTHON(self) -> Optional[str]:
        return environ.get('PYSPARK_PYTHON')

    @PYSPARK_PYTHON.setter
    def PYSPARK_PYTHON(self, value: Union[str, bytes, PathLike]) -> None:
        check_type_compatibility(value, (str, bytes, PathLike))
        environ['PYSPARK_PYTHON'] = str(value)


GlobalSparkConfig = _GlobalSparkConfigType()
